from typing import List
from ..asyncpy import HTTPLoop
from ._language import Language
from ._engine import Engine


class BaseTranslator:
    def __init__(self, source: Language, dest: Language):
        self.src = source
        self.dst = dest
        self.__loop__ = HTTPLoop()

    def __a_translate_one__(self, text: str):
        raise NotImplementedError()

    def translate_n(self, texts: List[str], callback: callable = None):
        self.__loop__.clear()
        for t in texts:
            self.__a_translate_one__(t)

        if callback:
            self.__loop__.yield_done(callback)
        else:
            return self.__post_process__(self.__loop__.wait_all_done())

    def print_all_engines(self):
        print(Engine.__members__)

    def translate(self, text: str):
        self.__loop__.clear()
        self.__a_translate_one__(text)
        return self.__post_process__(self.__loop__.wait_all_done())

    def __post_process__(self, results):
        raise NotImplementedError()


class OpenAI_Translator(BaseTranslator):
    def __init__(self, source: Language, dest: Language, API_key):
        super().__init__(source, dest)
        self.api_key = API_key

    def __a_translate_one__(self, text: str, ):
        self.__loop__.post(
            url='https://api.openai.com/v1/chat/completions',
            headers={
                "Content-Type": "application/json",
                "Authorization": "Bearer {}".format(self.api_key)
            },
            json=dict(
                model='gpt-3.5-turbo',
                messages=[
                    {'role': 'system',
                     'content': f"Please translate the following sentences into {self.dst}. Only send back the translated texts and keep the punctuation, the special char such as \"[\", \"]\" unchanged. The source sentence is: "},
                    {"role": "user", "content": text},
                ],
                temperature=0
            )
        )

    def __post_process__(self, results):
        return [i['choices'][0]['message']['content'] for i in results]



from ._weather_query import weather_query as today_weather
from ._thesis_info import thesis_info as crawl_thesis_info
from ._ytb_channel_info import YTBInfoBot
from ._academic import PaperWithCodeQuery, ArxivQuery, IEEEQuery
from ._air_quality import AirQualityQuery
from ._exchange_rate_cn import exchange_rate_cn
from ._realpython_news import RealPythonNews
from ._abstracts_arxiv import query_abstracts_arxiv
from . import openpy
from . import asyncpy


import requests
from lxml import etree

headers = {
  'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.54 Safari/537.36',
}

def thesis_info(**kwargs):
    if 'urls' in kwargs:
        urls = kwargs['urls']
    else:
        urls = urls = [
            'https://wwwmatthes.in.tum.de/pages/kr45xfyxvcda/Anum-Afzal',
            'https://wwwmatthes.in.tum.de/pages/zvs4gw49ru7f/Phillip-Schneider',
            'https://wwwmatthes.in.tum.de/pages/1ejjrtvitero5/Stephen-Meisenbacher',
            'https://wwwmatthes.in.tum.de/pages/1n0vb1ew7e6u5/Tim-Schopf',
            ]

    result = ''

    for i in urls:
        html = requests.get(i,headers=headers).text
        content = ', '.join(etree.HTML(html).xpath('//span[@style]//text()')).strip()
        result += f'link: {i}, content: {content}'
        result += '\n'

    return result




import requests
from lxml import etree

headers = {
  'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.54 Safari/537.36',
}

def weather_query(city):
    html = requests.get(f'http://www.tianqi.com/{city.lower()}/', headers=headers).text

    time = etree.HTML(html).xpath('//dl[@class="weather_info"]//dd[@class="week"]//text()')[0].replace('\u3000',                                                                                                   ', ').strip()
    city = etree.HTML(html).xpath('//dl[@class="weather_info"]//dd[@class="name"]//text()')[0]
    weather = ', '.join(etree.HTML(html).xpath('//dl[@class="weather_info"]//dd[@class="weather"]//span//text()'))
    shidu = ', '.join(etree.HTML(html).xpath('//dl[@class="weather_info"]//dd[@class="shidu"]//text()'))
    kongqi = ', '.join(etree.HTML(html).xpath('//dl[@class="weather_info"]//dd[@class="kongqi"]//text()'))
    general_weather_info = f'{time}. {city}为{weather}, {shidu}, {kongqi}.'

    detailed_weather_info = ''
    for i in etree.HTML(html).xpath('//div[@class="mainWeather"]//ul[@class="raweather760"]//a/@title')[0:-1:2]:
        detailed_weather_info += i + '; '

    weather_info = f'''{general_weather_info}

    局部地区详细天气情况：
    {detailed_weather_info}'''

    return weather_info


# source: https://platform.openai.com/docs/api-reference/introduction
from pprint import pprint
import requests
from typing import Union, List
import aiohttp
import asyncio
from .config import ChatCompletionConfig
from ._return import BaseReturn,AsyncReturn
from openai.error import InvalidRequestError


class ChatGPT:
    def __init__(self, api_key:str, model:str='gpt-3.5-turbo', global_system:str= '', use_assist:bool = True):
        '''

        :param api_key: your openai api key
        :param model: model name
        :param global_system: The local_system message helps set the behavior of the assistant. In the example above, the assistant was instructed with “You are a helpful assistant.”
        '''
        self.model = model
        self.global_system = global_system
        # self.api_key = api_key
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(api_key)
        }
        self.knowledge_base= dict()
        self.use_assist = use_assist
        self.__namer__ = -999999

    def __get_unique__name__(self):
        self.__namer__+=1
        return str(self.__namer__)

    def print_KB(self):
        pprint(self.knowledge_base)

    def __update_KB__(self, key:str, value:str):
        if key not in self.knowledge_base:
            self.knowledge_base[key] = [value]
        else:
            self.knowledge_base[key].append(value)


    def __render_KB__(self,user_name:str):
        if user_name in self.knowledge_base:
            return ', '.join(self.knowledge_base[user_name])
        else:
            return ''
    def __build_headers_data__(self,param:ChatCompletionConfig):
        # 设置请求头和参数
        headers = self.headers
        user_msg = param.user_msg
        local_system = param.local_system
        temperature = param.temparature
        top_p = param.top_p
        n = param.n
        stream = param.stream
        stop = param.stop
        max_tokens = param.max_tokens
        presence_penalty = param.presence_penalty
        frequency_penalty = param.frequency_penalty
        user_name = param.user_name if param.user_name is not None else self.__get_unique__name__()

        if param.assistant is not None:
            assistant = param.assistant
        else:
            if self.use_assist:
                assistant = self.__render_KB__(param.user_name)
            else:
                assistant = ''

        data = dict(
            model=self.model,
            messages=[
                {'role': 'system', 'content': self.global_system if local_system is None else local_system},
                {"role": "user", "content": user_msg},
                {'role': "assistant", 'content': assistant}
            ],
            temperature=temperature,
            top_p=top_p,
            n=n,
            stream=stream,
            stop=stop,
            max_tokens=max_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            user=user_name
        )
        return dict(
            headers = headers,
            data = data
        )

    def reply(self,
                param: ChatCompletionConfig,
              ) ->Union[BaseReturn, AsyncReturn]:

        # 1. build headers and data
        _  = self.__build_headers_data__(param)
        headers = _['headers']
        data = _['data']

        # 2. update knowledge base
        self.__update_KB__(param.user_name,param.user_msg)

        # 3. return request
        return self.__request__(headers,data,param.only_response)

    def __request__(self,headers, data, only_response):
        raise NotImplementedError()


class SyncChatGPT(ChatGPT):

    def __request__(self, headers, data,only_response,jupyter=None):
        # send request
        print('send request...')
        response = requests.post("https://api.openai.com/v1/chat/completions", headers=headers, json=data)

        # process response
        print('receive response...')
        if response.status_code == 200:
            result = response.json()
            if only_response:
                tmp = result['choices']
                return BaseReturn([i['message']['content'] for i in tmp])

            return BaseReturn(result)
        elif response.status_code == 400:
            raise InvalidRequestError(message=data['messages'],param=data)
        else:
            raise IOError(response.json())


class AsyncChatGPT(ChatGPT):

    def __request__(self, headers, data, only_response):
        async def request(headers,data,only_response):
            url = 'https://api.openai.com/v1/chat/completions'
            print('send request...')
            async with aiohttp.ClientSession(headers=headers) as session:
                async with session.post(url, json=data) as response:
                    if response.status == 200:
                        result = await response.json()
                        print('receive response...')
                        if only_response:
                            tmp = result['choices']
                            return [i['message']['content'] for i in tmp]
                        return result
                    elif response.status == 400:
                        raise InvalidRequestError(message=data['messages'],param=data)
                    else:
                        raise IOError(response.json())

        return AsyncReturn(asyncio.get_event_loop().create_task(request(headers,data,only_response)))



    def multi_reply(self,
                            params: List[ChatCompletionConfig],
                          ):
        async def reply(param:ChatCompletionConfig):
            _ = self.__build_headers_data__(param)
            headers = _['headers']
            data = _['data']
            self.__update_KB__(param.user_name,param.user_msg)
            return await self.__request__(headers,data,param.only_response)

        async def m_request(params):
            return await asyncio.gather(*tuple(reply(param) for param in params))

        return AsyncReturn(asyncio.get_event_loop().create_task(m_request(params)))








import asyncio
import random
import threading
from asyncio import Task
from typing import List, Union,Dict,Callable


class BasisLoop:
    def __init__(self):
        self.__loop__ = asyncio.get_event_loop()
        self.__tasks__ = []

    def get_all_tasks(self):
        return self.__tasks__
    def get_done_tasks(self):
        return list(filter(lambda x: x.done(), self.__tasks__))
    def get_cancelled_tasks(self):
        return list(filter(lambda x: x.cancelled(), self.__tasks__))
    def get_pending_tasks(self):
        return list(filter(lambda x:not x.cancelled() and not x.done(), self.__tasks__))
    def wait_all_done(self):
        return self.join()
    def join(self):
        return [self.__loop__.run_until_complete(t) for t in self.__tasks__]
    def clear(self):
        self.__tasks__.clear()
    def get_task_statues(self):
        return [t._state for t in self.__tasks__]

    @staticmethod
    def __create_coro__(async_func, *args):
        return async_func(*args)

class PendingLoop(BasisLoop):
    '''
    example usage:
    asyncpy def dumi():
        print(0)
        await asyncio.sleep(1)
        print(1)


    loop = PendingLoop()
    loop.pend_new_task(dumi()) # task not started
    loop.pend_new_task(dumi()) # task not started
    loop.pend_new_task(dumi()) # task not started
    loop.wait_all_done()
    '''
    def __init__(self):
        super().__init__()

    def pend_new_task(self, coro, name=None) ->Task:
        task = self.__loop__.create_task(coro)
        if name is not None:
            task.set_name(name)
        self.__tasks__.append(task)
        return task

    def __pend_dumi_tasks__(self, n_tasks: int) -> List[Task]:
        async def dumi():
            print('task started')
            t = random.randint(0,10)
            await asyncio.sleep(t)
            print(f'task finished. Use time {t} seconds.')

        return [self.pend_new_task(dumi()) for _ in range(n_tasks)]


class ThreadPoolLoop(PendingLoop):
    '''
    example usage:

    import random
    asyncpy def dumi():
        print('task started')
        t = random.randint(0, 10)
        await asyncio.sleep(t)
        print(f'task finished. Use time {t} seconds.')

    loop = ThreadPoolLoop()
    loop.start_new_task(dumi) # task started
    loop.start_new_task(dumi)  # task started
    loop.start_new_task(dumi)  # task started

    loop.yield_done(print)
    '''
    def __init__(self):
        super().__init__()

    def start_new_task(self, async_func, *func_args, task_name:str=None):
        async def afun_ext():
            print(f"Thread ID: {threading.current_thread().ident}, Thread Name: {threading.current_thread().name}")
            return await async_func(*func_args)

        async def create_task():
            # coro = await self.__loop__.run_in_executor(None, self.__create_coro__, afun_ext) # 验证只有一个线程
            coro = await self.__loop__.run_in_executor(None, self.__create_coro__, async_func,*func_args)
            task = self.__loop__.create_task(coro)
            if task_name is not None:
                task.__name__ = task_name
            else:
                task.__name__ = f'Task {len(self.__tasks__)}'

            self.__tasks__.append(task)
            return task
        return self.__loop__.run_until_complete(create_task())

    async def __yield_done_generator__(self):
        """Async generator that yields completed tasks from a list of tasks."""
        completed = set()
        tasks = self.__tasks__
        while tasks:
            done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            for task in done - completed:
                completed.add(task)
                yield task

    def yield_done(self, callback:Union[Callable,Dict]):
        async def wrapper():
            async for t in self.__yield_done_generator__():
                print(f"{t.__name__} finished.")
                if isinstance(callback,dict):
                    callback_ = callback[t.__name__]
                    callback_(t.result())
                else:
                    callback(t.result())

        task= self.__loop__.create_task(wrapper())
        self.__loop__.run_until_complete(task)













from typing import Awaitable
import asyncio
from ..utils import in_jupyter_notebook

class BaseReturn:
    def __init__(self,data):
        '''
        Only accept key word arguments and store every argument as an attribute
        :param kwargs:
        '''
        self.data = data

    def eval(self):
        '''
        evaluate self
        :return:
        '''
        return self.data

    # subscriptable and slice-able
    def __getitem__(self, idx):
        return self.eval()[idx]

    # return an iterator that can be used in for loop etc.
    def __iter__(self):
        return self.eval().__iter__()

    def __len__(self):
        return len(self.eval())

    def __eq__(self, other):
        return self.eval().__eq__(other)

    def __str__(self) -> str:
        return str(self.eval())

    def __repr__(self):
        return repr(self.eval())


class AsyncReturn(BaseReturn, Awaitable):
    def __init__(self, task:Awaitable):
        super().__init__(task)
        self.value = None

    def eval(self):
        if self.value is not None:
            return self.value

        if in_jupyter_notebook():
            self.value = self.data  # jupyter: need await
        else:
            self.value = asyncio.get_event_loop().run_until_complete(self.data)  # script
        return self.value

    def __await__(self):
        return self.data.__await__()



from .loop import ThreadPoolLoop
import aiohttp

class HTTPLoop(ThreadPoolLoop):
    @classmethod
    async def __process_response__(cls,response):
        content_type = response.headers.get('Content-Type')
        if content_type and 'application/json' in content_type:
            # 响应是JSON数据
            data = await response.json()
            # 对数据进行处理
            return data

        elif content_type and 'text/html' in content_type:
            # 响应是HTML文档
            text = await response.text()
            return text

        else:
            # 响应类型未知
            # 处理错误
            # ...
            return IOError("Unknown content type")

    @classmethod
    async def __GET__(cls,url,headers,params,timeout,ssl):
        async with aiohttp.ClientSession() as session:
            async with session.get(url=url,headers=headers,params=params,timeout=timeout,ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __POST__(cls,url, headers, data,json, timeout, ssl):
        async with aiohttp.ClientSession() as session:
            async with session.post(url=url, headers=headers, data=data,json=json , timeout=timeout, ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __PUT__(cls, url, headers=None, data=None, timeout=10, ssl=None):
        async with aiohttp.ClientSession() as session:
            async with session.put(url=url, headers=headers, data=data, timeout=timeout,
                                   ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __PATCH__(cls, url, headers=None, data=None, timeout=10, ssl=None):
        async with aiohttp.ClientSession() as session:
            async with session.patch(url=url, headers=headers, data=data, timeout=timeout,
                                     ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __DELETE__(cls, url, headers=None, timeout=10, ssl=None):
        async with aiohttp.ClientSession() as session:
            async with session.delete(url=url, headers=headers, timeout=timeout,
                                      ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __HEAD__(cls, url,allow_redirects:bool=False ,headers=None, timeout=10, ssl=None):
        async with aiohttp.ClientSession() as session:
            async with session.head(url=url, headers=headers,allow_redirects=allow_redirects, timeout=timeout,
                                    ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    @classmethod
    async def __OPTIONS__(cls, url, headers=None,allow_redirects:bool=False , timeout=10, ssl=None):
        async with aiohttp.ClientSession() as session:
            async with session.options(url=url, headers=headers, allow_redirects=allow_redirects, timeout=timeout,
                                       ssl=ssl) as response:
                result = await cls.__process_response__(response)
                if response.status == 200:
                    return result
                else:
                    return IOError(result)

    def get(self,url,headers=None,params=None,timeout=None,ssl=False):
        self.start_new_task(self.__GET__,url,headers,params,timeout,ssl)

    def post(self,url,headers=None,data=None,json=None, timeout=None,ssl=False):
        self.start_new_task(self.__POST__,url,headers,data,json,timeout,ssl)

    def put(self, url, headers=None, data=None, timeout=None, ssl=False):
        self.start_new_task(self.__PUT__, url, headers, data, timeout, ssl)

    def patch(self, url, headers=None, data=None, timeout=None, ssl=False):
        self.start_new_task(self.__PATCH__, url, headers, data, timeout, ssl)

    def delete(self, url, headers=None, timeout=None, ssl=False):
        self.start_new_task(self.__DELETE__, url, headers, timeout, ssl)

    def head(self, url, headers=None, allow_redirects:bool=False , timeout=None, ssl=False):
        self.start_new_task(self.__HEAD__, url, allow_redirects, headers, timeout, ssl)

    def options(self, url, headers=None, allow_redirects:bool=False , timeout=None, ssl=False):
        self.start_new_task(self.__OPTIONS__, url, headers, allow_redirects, timeout, ssl)






import requests
from lxml import etree

class RealPythonNews:
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.54 Safari/537.36',
    }

    ENDPOINT = 'https://realpython.com'

    @classmethod
    def all(cls, return_links=False):
        '''
        return all news titles and additionally the links if return_links = True
        '''
        html = requests.get(cls.ENDPOINT + '/tutorials/community', headers=cls.headers).text
        titles = etree.HTML(html).xpath('//h2[@class="card-title h4 my-0 py-0"]//text()')
        if return_links:
            hrefs = etree.HTML(html).xpath('//a[@class=" "]/@href')
            hrefs = [cls.ENDPOINT + x for x in hrefs]
            return titles,hrefs
        return titles

    @classmethod
    def latests(cls, return_links = False):
        if return_links:
            tmp = cls.all(return_links)
            return tmp[0][0], tmp[1][0]
        else:
            return cls.all()[0]




if __name__ == '__main__':
    print(RealPythonNews.latests(True))
    print(RealPythonNews.latests())




from typing import Union, List


class ChatCompletionConfig:
    def __init__(self,*,
        user_msg: str,
        user_name: str = None,
        assistant: str = None,
        local_system:str = None,
        temperature:float = 1,
        top_p:float = 1,
        n:int = 1,
        stream:bool = False,
        stop: Union[str, List[str]] = None,
        max_tokens:int = 1000,
        presence_penalty:float = 0,
        frequency_penalty:float = 0,
        only_response = True,
                 ):
        '''

         :param user_msg: user message
         :param assistant: external knowledge base. E.g. chat history
         :param temperature: What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the     output more random, while lower values like 0.2 will make it more focused and deterministic.
             We generally recommend altering this or top_p but not both.
         :param top_p: An alternative to sampling with temperature, called nucleus sampling, where the model         considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising    the top 10% probability mass are considered.
             We generally recommend altering this or temperature but not both.
         :param n: How many chat completion choices to generate for each input message.
         :param stream: If set, partial message deltas will be sent, like in ChatGPT. Tokens will be sent as data-only server-sent events as they become available, with the stream terminated by a data: [DONE] message.
         :param stop: Up to 4 sequences where the API will stop generating further tokens.
         :param max_tokens: The maximum number of tokens allowed for the generated answer. By default, the number of tokens the model can return will be (4096 - prompt tokens).
         :param presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.
         :param frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.
         :param user_name: A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. Learn more.
         :param only_response: if only return the text response from ChatGPT

         :return:
         '''
        self.user_name = user_name
        self.user_msg = user_msg
        self.assistant = assistant
        self.local_system =  local_system
        self.temparature = temperature
        self.top_p = top_p
        self.n = n
        self.stream = stream
        self.stop = stop
        self.max_tokens = max_tokens
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.only_response = only_response

    def to_dict(self) -> dict:
        ret = {}
        for key, value in self.__dict__.items():
            if hasattr(value, 'to_dict'):
                ret[key] = value.to_dict()
            else:
                ret[key] = value
        return ret


def  in_jupyter_notebook():
    # 定义一个变量，表示是否在 ipython 会话中
    _in_ipython_session = False
    try:
        # 尝试获取 __IPYTHON__ 这个变量
        __IPYTHON__
        # 如果没有异常，说明在 ipython 会话中
        _in_ipython_session = True
    except NameError:
        # 如果出现异常，说明不在 ipython 会话中
        _in_ipython_session = False

    # 打印结果
    return _in_ipython_session